!2-port network, S-parameter and noise data
!Default MA format, GHz frequencies, 50 ohm reference, S-parameters
[Version] 2.0
#
[Number of Ports] 2
[Two-Port Data Order] 21_12
[Number of Frequencies] 2
[Number of Noise Frequencies] 2 
[Network Data]
1 111 0 121 0 112 0 122 0
2 211 0 221 0 212 0 222 0