!1-port Z-parameter file, multiple frequency points
[Version] 2.0
# MHz Z MA
[Number of Ports] 1
[Number of Frequencies] 5
[Reference] 20.0
[Network Data]
!freq magZ11 angZ11
100 74.25 -4
200 60 -22
300 53.025 -45
400 30 -62
500 0.75 -89 