!2-port H-parameter file, single frequency point
[Version] 2.0 ! It looks like the version string is wrong in specification
# kHz G MA R 1
[Number of Ports] 2
[Two-Port Data Order] 21_12
[Number of Frequencies] 1
[Matrix Format] Full
[Network Data]
! freq magH11 angH11 magH21 angH21 magH12 angH12 magH22 angH22
2 1.03818 13.03373 5.61562 4.03729 0.06292 -76.627 1.49435 1.03729